import logging
from typing import TypeAlias, cast

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import EdgeApply, NodeApply
from pydantic_core import ErrorDetails
from rdflib.term import Node

from cognite.neat.exceptions import NeatException
from cognite.neat.graph.stores.graph_store import NeatGraphStore
from cognite.neat.rules.exporter._rules2dms import DataModel
from cognite.neat.rules.exporter._rules2pydantic_models import add_class_prefix_to_xid, rules_to_pydantic_models
from cognite.neat.rules.models.rules import Rules
from cognite.neat.utils.utils import chunker, datetime_utc_now, retry_decorator

Triple: TypeAlias = tuple[Node, Node, Node]


def rdf2nodes_and_edges(
    graph_store: NeatGraphStore,
    transformation_rules: Rules,
    stop_on_exception: bool = False,
    add_class_prefix: bool = False,
) -> tuple[list[NodeApply], list[EdgeApply], list[ErrorDetails]]:
    """Generates DMS nodes and edges from knowledge graph stored as RDF triples

    Args:
        graph_store: Instance of NeatGraphStore holding RDF graph
        transformation_rules: Transformation rules holding data model definition
        stop_on_exception: Whether to stop execution on exception. Defaults to False.
        add_class_prefix: Whether to add class name as a prefix to instance external id. Defaults to False.

    Returns:
        Tuple holding nodes, edges and exceptions
    """
    if transformation_rules.metadata.namespace is None:
        raise ValueError("Namespace is not defined in transformation rules metadata")

    nodes: list[NodeApply] = []
    edges: list[EdgeApply] = []
    exceptions: list[ErrorDetails] = []

    data_model = DataModel.from_rules(transformation_rules)
    pydantic_models = rules_to_pydantic_models(transformation_rules)

    for class_ in transformation_rules.classes:
        if class_ in data_model.containers:
            class_namespace = transformation_rules.metadata.namespace[class_]
            class_instance_ids = [
                cast(Triple, res)[0]
                for res in graph_store.query(f"SELECT ?instance WHERE {{ ?instance rdf:type <{class_namespace}> . }}")
            ]

            counter = 0
            start_time = datetime_utc_now()
            total = len(class_instance_ids)

            for class_instance_id in class_instance_ids:
                counter += 1
                try:
                    instance = pydantic_models[class_].from_graph(  # type: ignore[attr-defined]
                        graph_store, transformation_rules, class_instance_id
                    )
                    if add_class_prefix:
                        instance.external_id = add_class_prefix_to_xid(
                            class_name=instance.__class__.__name__, external_id=instance.external_id
                        )
                    new_node = instance.to_node(data_model, add_class_prefix)
                    is_valid, reason = is_node_valid(new_node)
                    if not is_valid:
                        exceptions.append(
                            ErrorDetails(
                                input=class_instance_id,
                                loc=tuple(["Nodes"]),
                                msg=f"Not valid node {new_node.external_id}. Reason: {reason}",
                                type="Node validation error",
                            )
                        )
                        continue
                    nodes.append(new_node)

                    new_edges = instance.to_edge(data_model, add_class_prefix)
                    for new_edge in new_edges:
                        is_valid, reason = is_edge_valid(new_edge)
                        if not is_valid:
                            exceptions.append(
                                ErrorDetails(
                                    input=class_instance_id,
                                    loc=tuple(["Edges"]),
                                    msg=f"Not valid edge {new_edge.external_id}. Reason: {reason}",
                                    type="Edge validation error",
                                )
                            )
                            continue
                        edges.append(new_edge)

                    delta_time = datetime_utc_now() - start_time
                    delta_time = (delta_time.seconds * 1000000 + delta_time.microseconds) / 1000
                    msg = (
                        f"{class_} {counter} of {total} instances processed, "
                        f"instance processing time: {delta_time/counter:.2f} "
                    )
                    msg += f"ms ETC: {(delta_time/counter) * (total - counter) / 1000 :.3f} s"
                    logging.info(msg)

                except Exception as e:
                    logging.error(
                        f"Instance {class_instance_id} of {class_} cannot be resolved to nodes and edges. Reason: {e}"
                    )
                    if stop_on_exception:
                        raise e

                    if isinstance(e, NeatException):
                        exceptions.append(e.to_error_dict())
                    else:
                        exceptions.append(
                            ErrorDetails(
                                input=class_instance_id,
                                loc=tuple(["rdf2nodes_and_edges"]),
                                msg=str(e),
                                type=f"Exception of type {type(e).__name__} occurred  \
                                when processing instance of {class_}",
                            )
                        )
                    continue

    return nodes, edges, exceptions


def is_node_valid(node: NodeApply) -> tuple[bool, str]:
    return is_valid_external_id(node.external_id)


def is_edge_valid(edge: EdgeApply) -> tuple[bool, str]:
    for external_id in [edge.external_id, edge.start_node.external_id, edge.end_node.external_id]:
        is_valid, reason = is_valid_external_id(external_id)
        if not is_valid:
            return False, reason
    return True, ""


def is_valid_external_id(external_id: str) -> tuple[bool, str]:
    if external_id is None or external_id == "" or len(external_id) >= 255:
        return False, f"external_id {external_id} is empty of too long"
    return True, ""


def upload_nodes(
    client: CogniteClient, nodes: list[NodeApply], batch_size: int = 1000, max_retries: int = 1, retry_delay: int = 3
):
    """Uploads nodes to CDF

    Args:
        client: Instance of CogniteClient
        nodes: List of nodes to upload to CDF
        batch_size: Size of batch. Defaults to 1000.
        max_retries: Maximum times to retry the upload. Defaults to 1.
        retry_delay: Time delay before retrying the upload. Defaults to 3.

    !!! note "batch_size"
        If batch size is set to 1 or None, all nodes will be pushed to CDF in one go.
    """
    if batch_size:
        logging.info(f"Uploading nodes in batches of {batch_size}")
        _micro_batch_push(client, nodes, batch_size, message="Upload", max_retries=max_retries, retry_delay=retry_delay)

    else:
        logging.info("Batch size not set, pushing all nodes to CDF in one go!")

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="create-nodes")
        def create_nodes():
            client.data_modeling.instances.apply(nodes=nodes)

        create_nodes()


def upload_edges(
    client: CogniteClient, edges: list[EdgeApply], batch_size: int = 5000, max_retries: int = 1, retry_delay: int = 3
):
    """Uploads edges to CDF

    Args:
        client: Instance of CogniteClient
        edges: List of edges to upload to CDF
        batch_size: Size of batch. Defaults to 5000.
        max_retries: Maximum times to retry the upload. Defaults to 1.
        retry_delay: Time delay before retrying the upload. Defaults to 3.

    !!! note "batch_size"
        If batch size is set to 1 or None, all edges will be pushed to CDF in one go.

    """
    if batch_size:
        logging.info(f"Uploading edges in batches of {batch_size}")
        _micro_batch_push(client, edges, batch_size, message="Upload", max_retries=max_retries, retry_delay=retry_delay)

    else:
        logging.info("Batch size not set, pushing all edges to CDF in one go!")

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="create-edges")
        def create_nodes():
            client.data_modeling.instances.apply(edges=edges, auto_create_start_nodes=True, auto_create_end_nodes=True)

        create_nodes()


def _micro_batch_push(
    client: CogniteClient,
    nodes_or_edges: list[NodeApply] | list[EdgeApply],
    batch_size: int = 1000,
    message: str = "Upload",
    max_retries: int = 1,
    retry_delay: int = 3,
):
    """Uploads nodes or edges in batches

    Args:
        client: Instance of CogniteClient
        nodes_or_edges: List of nodes or edges
        batch_size: Size of batch. Defaults to 1000.
        push_type: Type of push, either "nodes" or "edges". Defaults to "nodes".
        message: Message to logged. Defaults to "Upload".
        max_retries: Maximum times to retry the upload. Defaults to 1.
        retry_delay: Time delay before retrying the upload. Defaults to 3.
    """
    total = len(nodes_or_edges)
    counter = 0

    if nodes_or_edges and isinstance(nodes_or_edges[0], NodeApply):
        push_type = "nodes"
    elif nodes_or_edges and isinstance(nodes_or_edges[0], EdgeApply):
        push_type = "edges"
    else:
        raise ValueError("nodes_or_edges must be a list of NodeApply or EdgeApply objects")

    for batch in chunker(nodes_or_edges, batch_size):
        counter += len(batch)
        start_time = datetime_utc_now()

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name=f"microbatch-{push_type}")
        def upsert_nodes_or_edges(upload_batch):
            if push_type == "nodes":
                client.data_modeling.instances.apply(nodes=upload_batch)
            elif push_type == "edges":
                client.data_modeling.instances.apply(
                    edges=upload_batch, auto_create_start_nodes=True, auto_create_end_nodes=True
                )

        upsert_nodes_or_edges(batch)

        delta_time = (datetime_utc_now() - start_time).seconds

        msg = f"{message} {counter} of {total} {push_type}, batch processing time: {delta_time:.2f} "
        msg += f"seconds ETC: {delta_time * (total - counter) / (60*batch_size) :.2f} minutes"
        logging.info(msg)



"""This module performs importing of various formats to one of serializations for which
there are loaders to TransformationRules pydantic class."""

# TODO: if this module grows too big, split it into several files and place under ./converter directory

from pathlib import Path

import numpy as np
import pandas as pd
from rdflib import DC, DCTERMS, OWL, RDF, RDFS, SKOS, Graph

from cognite.neat.rules.importer._base import BaseImporter
from cognite.neat.rules.models.tables import Tables
from cognite.neat.utils.utils import get_namespace, remove_namespace


class OWLImporter(BaseImporter):
    """Convert OWL ontology to tables/ transformation rules / Excel file.

        Args:
            owl_filepath: Path to OWL ontology

    !!! Note
        OWL Ontologies typically lacks some information that is required for making a complete
        data model. This means that the methods .to_rules() will typically fail. Instead, it is recommended
        that you use the .to_spreadsheet() method to generate an Excel file, and then manually add the missing
        information to the Excel file. The Excel file can then be converted to a TransformationRules object.

        One can set the `use_default_values` parameter to True to allow neat to set default
        values for missing information.

    """

    def __init__(self, owl_filepath: Path):
        self.owl_filepath = owl_filepath

    def to_tables(self, use_default_values: bool = False) -> dict[str, pd.DataFrame]:
        graph = Graph()
        try:
            graph.parse(self.owl_filepath)
        except Exception as e:
            raise Exception(f"Could not parse owl file: {e}") from e

        # bind key namespaces
        graph.bind("owl", OWL)
        graph.bind("rdf", RDF)
        graph.bind("rdfs", RDFS)
        graph.bind("dcterms", DCTERMS)
        graph.bind("dc", DC)
        graph.bind("skos", SKOS)

        return {
            Tables.metadata: _parse_owl_metadata_df(graph, use_default_values=use_default_values),
            Tables.classes: _parse_owl_classes_df(graph, use_default_values=use_default_values),
            Tables.properties: _parse_owl_properties_df(graph, use_default_values=use_default_values),
        }


def _create_default_metadata_parsing_config() -> dict[str, tuple[str, ...]]:
    # TODO: these are to be read from Metadata pydantic model
    return {
        "header": (
            "namespace",
            "prefix",
            "dataModelName",
            "cdfSpaceName",
            "version",
            "isCurrentVersion",
            "created",
            "updated",
            "title",
            "description",
            "creator",
            "contributor",
            "rights",
            "license",
        )
    }


def _create_default_classes_parsing_config() -> dict[str, tuple[str, ...]]:
    # TODO: these are to be read from Class pydantic model
    return {
        "header": (
            "Class",
            "Name",
            "Description",
            "Parent Class",
            "Deprecated",
            "Deprecation Date",
            "Replaced By",
            "Source",
            "Source Entity Name",
            "Match",
            "Comment",
        )
    }


def _create_default_properties_parsing_config() -> dict[str, tuple[str, ...]]:
    # TODO: these are to be read from Property pydantic model
    return {
        "header": (
            "Class",
            "Property",
            "Name",
            "Description",
            "Type",
            "Min Count",
            "Max Count",
            "Deprecated",
            "Deprecation Date",
            "Replaced By",
            "Source",
            "Source Entity Name",
            "Match",
            "Comment",
        )
    }


def _parse_owl_metadata_df(
    graph: Graph, parsing_config: dict | None = None, use_default_values: bool = False
) -> pd.DataFrame:
    """Parse owl metadata from graph to pandas dataframe.

    Args:
        graph: Graph containing owl metadata
        parsing_config: Configuration for parsing. Defaults to None.

    Returns:
        Dataframe containing owl metadata
    """
    if parsing_config is None:
        parsing_config = _create_default_metadata_parsing_config()

    if use_default_values:
        query = (
            'SELECT ?namespace (COALESCE(?prefix, "neat") AS ?prefix)'
            ' (COALESCE(?dataModelName, "neat") AS ?dataModelName)'
            ' (COALESCE(?cdfSpaceName, "playground") AS ?cdfSpaceName)'
            ' (COALESCE(?version, "1.0.0") AS ?version)'
            ' (COALESCE(?isCurrentVersion, "true"^^xsd:boolean) AS ?isCurrentVersion)'
            ' (COALESCE(?created, "1983-01-22T02:00:00Z"^^xsd:dateTime) AS ?created)'
            ' (COALESCE(?updated, "2021-11-13T00:00:00Z"^^xsd:dateTime) AS ?updated)'
            ' (COALESCE(?title, "OWL Inferred Data Model") AS ?title)'
            ' (COALESCE(?creator, "NEAT") AS ?creator)'
            ' (COALESCE(?contributor, "NEAT") AS ?contributor)'
            ' (COALESCE(?description, "This data model has been inferred with NEAT") AS ?description)'
            ' (COALESCE(?rights, "Unknown rights of usage") AS ?rights)'
            ' (COALESCE(?license, "Unknown license") AS ?license)'
        )
    else:
        query = (
            "SELECT ?namespace ?prefix ?dataModelName ?cdfSpaceName ?version ?isCurrentVersion "
            "?created ?updated ?title ?description ?creator ?contributor ?rights ?license"
        )

    query += """
    WHERE {
        ?namespace a owl:Ontology .
        OPTIONAL {?namespace owl:versionInfo ?version }.
        OPTIONAL {?namespace dcterms:creator ?creator }.
        OPTIONAL {?namespace dcterms:title|rdfs:label|skos:prefLabel ?title }.
        OPTIONAL {?namespace dcterms:contributor ?contributor }.
        OPTIONAL {?namespace dcterms:modified ?updated }.
        OPTIONAL {?namespace dcterms:created ?created }.
        OPTIONAL {?namespace dcterms:description ?description }.

        OPTIONAL {?namespace dcterms:rights|dc:rights ?rights }.

        OPTIONAL {?namespace dcterms:license|dc:license ?license }.
        FILTER (!isBlank(?namespace))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
        FILTER (!bound(?title) || LANG(?title) = "" || LANGMATCHES(LANG(?title), "en"))
    }
    """

    raw_df = pd.DataFrame(list(graph.query(query)), columns=parsing_config["header"])
    if raw_df.empty:
        return raw_df
    raw_df.replace(np.nan, "", regex=True, inplace=True)

    clean_list = [
        [
            raw_df.namespace.unique()[0],
            "neat",
            raw_df.dataModelName.unique()[0],
            raw_df.cdfSpaceName.unique()[0],
            raw_df.version.unique()[0],
            True,
            raw_df.created.unique()[0],
            raw_df.updated.unique()[0],
            raw_df.title.unique()[0],
            raw_df.description.unique()[0],
            ", ".join(list(raw_df.creator.unique())),
            ", ".join(list(raw_df.contributor.unique())),
            raw_df.rights.unique()[0],
            raw_df.license.unique()[0],
        ]
    ]

    df = pd.DataFrame(np.vstack((parsing_config["header"], clean_list)).T)
    df.fillna("", inplace=True)

    return df


def _parse_owl_classes_df(
    graph: Graph, parsing_config: dict | None = None, use_default_values: bool = False
) -> pd.DataFrame:
    """Get all classes from the graph and their parent classes.

    Parameters
    ----------
    graph : Graph
        Graph to query
    parsing_config : dict, optional
        Configuration for parsing the dataframe, by default None

    Returns
    -------
    pd.DataFrame
        Dataframe with columns: class, parentClass
    """
    if parsing_config is None:
        parsing_config = _create_default_classes_parsing_config()

    query = """
SELECT ?class ?name ?description ?parentClass ?deprecated ?deprecationDate
?replacedBy ?source ?sourceEntity ?match ?comment
    WHERE {
        ?class a owl:Class .
        OPTIONAL {?class rdfs:subClassOf ?parentClass }.
        OPTIONAL {?class rdfs:label ?name }.
        OPTIONAL {?class rdfs:comment ?description} .
        OPTIONAL {?class owl:deprecated ?deprecated} .
        FILTER (!isBlank(?class))
        FILTER (!bound(?parentClass) || !isBlank(?parentClass))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
    }
    """

    raw_df = pd.DataFrame(list(graph.query(query)), columns=parsing_config["header"])
    if raw_df.empty:
        raw_df = pd.concat([raw_df, pd.DataFrame([len(raw_df) * [""]])], ignore_index=True)
        return raw_df
    raw_df.replace(np.nan, "", regex=True, inplace=True)

    raw_df.Source = raw_df.Class.apply(lambda x: get_namespace(x))
    raw_df.Class = raw_df.Class.apply(lambda x: remove_namespace(x))
    raw_df["Source Entity Name"] = raw_df.Class
    raw_df["Match"] = len(raw_df) * ["exact"]
    raw_df["Parent Class"] = raw_df["Parent Class"].apply(lambda x: remove_namespace(x))

    grouped_df = raw_df.groupby("Class")

    clean_list = [
        [
            class_,
            group_df["Name"].unique()[0],
            "\n".join(list(group_df.Description.unique())),
            ", ".join(list(group_df["Parent Class"].unique())),
            group_df.Deprecated.unique()[0],
            group_df["Deprecation Date"].unique()[0],
            group_df["Replaced By"].unique()[0],
            group_df["Source"].unique()[0],
            group_df["Source Entity Name"].unique()[0],
            group_df["Match"].unique()[0],
            group_df["Comment"].unique()[0],
        ]
        for class_, group_df in grouped_df
    ]

    df = pd.DataFrame(columns=parsing_config["header"], data=clean_list)
    df.fillna("", inplace=True)

    return df


def _parse_owl_properties_df(
    graph: Graph, parsing_config: dict | None = None, use_default_values: bool = False
) -> pd.DataFrame:
    """Get all properties from the OWL ontology

    Parameters
    ----------
    graph : Graph
        Graph to query
    parsing_config : dict, optional
        Configuration for parsing the dataframe, by default None

    Returns
    -------
    pd.DataFrame
        Dataframe with columns: class, property, name, ...
    """
    if parsing_config is None:
        parsing_config = _create_default_properties_parsing_config()

    query = """
    SELECT ?class ?property ?name ?description ?type ?minCount ?maxCount
    ?deprecated ?deprecationDate ?replacedBy ?source ?sourceEntity ?match ?comment
    WHERE {
        ?property a ?propertyType.
        FILTER (?propertyType IN (owl:ObjectProperty, owl:DatatypeProperty ) )
        OPTIONAL {?property rdfs:domain ?class }.
        OPTIONAL {?property rdfs:range ?type }.
        OPTIONAL {?property rdfs:label ?name }.
        OPTIONAL {?property rdfs:comment ?description} .
        FILTER (!isBlank(?property))
        FILTER (!bound(?class) || !isBlank(?class))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
        OPTIONAL {?property owl:deprecated ?deprecated} .
    }
    """

    raw_df = pd.DataFrame(list(graph.query(query)), columns=parsing_config["header"])
    if raw_df.empty:
        return raw_df
    raw_df.replace(np.nan, "", regex=True, inplace=True)

    raw_df.Source = raw_df.Property.apply(lambda x: get_namespace(x))
    raw_df.Class = raw_df.Class.apply(lambda x: remove_namespace(x))
    raw_df.Property = raw_df.Property.apply(lambda x: remove_namespace(x))
    raw_df.Type = raw_df.Type.apply(lambda x: remove_namespace(x))
    raw_df["Source Entity Name"] = raw_df.Property
    raw_df["Match"] = len(raw_df) * ["exact"]

    class_grouped_dfs = raw_df.groupby("Class")

    clean_list = []

    for class_, class_grouped_df in class_grouped_dfs:
        property_grouped_dfs = class_grouped_df.groupby("Property")
        for property_, property_grouped_df in property_grouped_dfs:
            clean_list += [
                [
                    class_,
                    property_,
                    property_grouped_df["Name"].unique()[0],
                    "\n".join(list(property_grouped_df.Description.unique())),
                    property_grouped_df.Type.unique()[0],
                    property_grouped_df["Min Count"].unique()[0],
                    property_grouped_df["Max Count"].unique()[0],
                    property_grouped_df.Deprecated.unique()[0],
                    property_grouped_df["Deprecation Date"].unique()[0],
                    property_grouped_df["Replaced By"].unique()[0],
                    property_grouped_df["Source"].unique()[0],
                    property_grouped_df["Source Entity Name"].unique()[0],
                    property_grouped_df["Match"].unique()[0],
                    property_grouped_df["Comment"].unique()[0],
                ]
            ]

    df = pd.DataFrame(columns=parsing_config["header"], data=clean_list)
    df.fillna("", inplace=True)

    return df



from pathlib import Path

import pandas as pd

from cognite.neat.rules.importer._base import BaseImporter


class XMLImporter(BaseImporter):
    """
    Importer for XML files.

    Args:
        xml_directory: Path to directory containing XML files.
    """

    def __init__(self, xml_directory: Path):
        self.xml_directory = xml_directory

    def to_tables(self) -> dict[str, pd.DataFrame]:
        raise NotImplementedError



import logging
import time
from pathlib import Path
from typing import ClassVar, cast

from prometheus_client import Gauge

from cognite.neat.rules import importer
from cognite.neat.utils.utils import generate_exception_report
from cognite.neat.workflows import utils
from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.cdf_store import CdfStore
from cognite.neat.workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat.workflows.steps.data_contracts import RulesData
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = ["LoadTransformationRules"]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()


class LoadTransformationRules(Step):
    """
    This step loads transformation rules from the file or remote location
    """

    description = "This step loads transformation rules from the file or remote location"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="validation_report_storage_dir",
            value="rules_validation_report",
            label="Directory to store validation report",
        ),
        Configurable(
            name="validation_report_file",
            value="rules_validation_report.txt",
            label="File name to store validation report",
        ),
        Configurable(
            name="file_name",
            value="rules.xlsx",
            label="Full name of the rules file in rules folder. If includes path, \
                it will be relative to the neat data folder",
        ),
        Configurable(name="version", value="", label="Optional version of the rules file"),
    ]

    def run(self, cdf_store: CdfStore) -> (FlowMessage, RulesData):  # type: ignore[syntax, override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        store = cdf_store
        # rules file
        if self.configs is None:
            raise ValueError(f"Step {type(self).__name__} has not been configured.")
        rules_file = Path(self.configs["file_name"])
        if str(rules_file.parent) == ".":
            rules_file_path = Path(self.data_store_path) / "rules" / rules_file
        else:
            rules_file_path = Path(self.data_store_path) / rules_file

        version = self.configs["version"]

        # rules validation
        report_file = self.configs["validation_report_file"]
        report_dir_str = self.configs["validation_report_storage_dir"]
        report_dir = self.data_store_path / Path(report_dir_str)
        report_dir.mkdir(parents=True, exist_ok=True)
        report_full_path = report_dir / report_file

        if not rules_file_path.exists():
            logging.info(f"Rules files doesn't exist in local fs {rules_file_path}")

        if rules_file_path.exists() and not version:
            logging.info(f"Loading rules from {rules_file_path}")
        elif rules_file_path.exists() and version:
            hash = utils.get_file_hash(rules_file_path)
            if hash != version:
                store.load_rules_file_from_cdf(str(rules_file), version)
        else:
            store.load_rules_file_from_cdf(str(rules_file), version)

        raw_rules = importer.ExcelImporter(rules_file_path).to_raw_rules()
        rules, errors, warnings_ = raw_rules.to_rules(return_report=True, skip_validation=False)
        report = generate_exception_report(errors, "Errors") + generate_exception_report(warnings_, "Warnings")

        with report_full_path.open(mode="w") as file:
            file.write(report)

        text_for_report = (
            "<p></p>"
            "Download rules validation report "
            f'<a href="http://localhost:8000/data/{report_dir_str}/{report_file}?{time.time()}" '
            f'target="_blank">here</a>'
        )

        if rules is None:
            return FlowMessage(
                error_text=f"Failed to load transformation rules! {text_for_report}",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        if self.metrics is None:
            raise ValueError(f"Step {type(self).__name__} has not been configured.")
        rules_metrics = cast(
            Gauge,
            self.metrics.register_metric(
                "data_model_rules", "Transformation rules stats", m_type="gauge", metric_labels=["component"]
            ),
        )
        rules_metrics.labels({"component": "classes"}).set(len(rules.classes))
        rules_metrics.labels({"component": "properties"}).set(len(rules.properties))
        logging.info(f"Loaded prefixes {rules.prefixes!s} rules from {rules_file_path.name!r}.")
        output_text = f"<p></p>Loaded {len(rules.properties)} rules! {text_for_report}"

        return FlowMessage(output_text=output_text), RulesData(rules=rules)



import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

from pydantic import BaseModel


class RdfStoreType(StrEnum):
    """RDF Store type"""

    MEMORY = "memory"
    FILE = "file"
    GRAPHDB = "graphdb"
    SPARQL = "sparql"
    OXIGRAPH = "oxigraph"


class RdfStoreConfig(BaseModel):
    type: RdfStoreType
    file_path: str | None = None
    query_url: str | None = None
    update_url: str | None = None
    user: str | None = None
    password: str | None = None
    api_root_url: str = "http://localhost:7200"



from pathlib import Path

from rdflib import Graph, Namespace

from cognite.neat.constants import DEFAULT_NAMESPACE, DEFAULT_URI, PREFIXES

__all__ = ["rdf_file_to_graph"]


def rdf_file_to_graph(
    filepath: Path,
    base_prefix: str = DEFAULT_URI,
    base_namespace: Namespace = DEFAULT_NAMESPACE,
    prefixes: dict[str, Namespace] = PREFIXES,
) -> Graph:
    """Created rdflib Graph instance loaded with RDF triples from file

    Args:
        filepath: Path to the RDF file
        base_prefix: base prefix for URIs. Defaults to DEFAULT_URI.
        base_namespace: base namespace for URIs . Defaults to DEFAULT_NAMESPACE.
        prefixes: Dictionary of prefixes to bind to graph. Defaults to PREFIXES.

    Returns:
        Graph instance loaded with RDF triples from file
    """
    graph = Graph()
    if filepath.is_file():
        graph.parse(filepath, publicID=base_namespace)
    else:
        for filename in filepath.iterdir():
            if filename.is_file():
                graph.parse(filename, publicID=base_namespace)

    graph.bind(base_prefix, base_namespace)
    if prefixes:
        for prefix, namespace in prefixes.items():
            graph.bind(prefix, namespace)

    return graph



"""Exports rules to CDF Data Model Storage (DMS) through cognite-sdk.
"""

import logging
import sys
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Literal, cast

import yaml

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import (
    ContainerApply,
    ContainerApplyList,
    ContainerId,
    ContainerProperty,
    DataModelApply,
    DataModelId,
    DirectRelation,
    DirectRelationReference,
    MappedPropertyApply,
    SpaceApply,
    ViewApply,
    ViewId,
)
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType
from cognite.client.data_classes.data_modeling.views import (
    ConnectionDefinitionApply,
    SingleHopConnectionDefinitionApply,
)
from pydantic import BaseModel, ConfigDict

from cognite.neat.rules import exceptions
from cognite.neat.rules.analysis import to_class_property_pairs
from cognite.neat.rules.exporter._base import BaseExporter
from cognite.neat.rules.exporter._validation import are_entity_names_dms_compliant, are_properties_redefined
from cognite.neat.rules.models.rules import Property, Rules
from cognite.neat.rules.type_mapping import DATA_TYPE_MAPPING
from cognite.neat.utils.utils import generate_exception_report


@dataclass
class DMSSchema:
    data_model: DataModelApply
    containers: ContainerApplyList


class DMSExporter(BaseExporter[DMSSchema]):
    """Class for exporting transformation rules object to CDF Data Model Storage (DMS).

    Args:
        rules: Transformation rules object.
        data_model_id: The id of the data model to be created.
        container_policy: How to create/reuse existing containers.
        existing_model: In the case of updating an existing model, this is the existing model.
        report: Report. This is used when the exporter object is created from RawRules
    """

    def __init__(
        self,
        rules: Rules,
        data_model_id: dm.DataModelId | None = None,
        container_policy: Literal["one-to-one-view", "extend-existing", "neat-optimized"] = "one-to-one-view",
        existing_model: dm.DataModel[dm.View] | None = None,
        report: str | None = None,
    ):
        super().__init__(rules, report)
        self.data_model_id = data_model_id
        self.container_policy = container_policy
        if container_policy == "extend-existing" and existing_model is None:
            raise ValueError("Container policy is extend-existing, but no existing model is provided")
        if container_policy != "one-to-one-view":
            raise NotImplementedError("Only one-to-one-view container policy is currently supported")
        self.existing_model = existing_model

    def _export_to_file(self, filepath: Path) -> None:
        if filepath.suffix not in {".yaml", ".yml"}:
            warnings.warn("File extension is not .yaml, adding it to the file name", stacklevel=2)
            filepath = filepath.with_suffix(".yaml")
        schema = self.export()
        filepath.write_text(
            yaml.safe_dump(
                {
                    "data_models": [schema.data_model.dump(camel_case=True)],
                    "containers": schema.containers.dump(camel_case=True),
                }
            )
        )

    def export(self) -> DMSSchema:
        model = DataModel.from_rules(self.rules, self.data_model_id)
        return DMSSchema(
            data_model=DataModelApply(
                space=model.space,
                external_id=model.external_id,
                version=model.version,
                description=model.description,
                name=model.name,
                views=list(model.views.values()),
            ),
            containers=ContainerApplyList(model.containers.values()),
        )


class DataModel(BaseModel):
    """
    Data model pydantic class used to create space, containers, views and data model in CDF.

    This can be used to create a data model in CDF from transformation rules.

    Args:
        space: Name of the space to place the resulting data model.
        external_id: External id of the data model.
        version: Version of the data model.
        description: Description of the data model.
        name: Name of the data model.
        containers: Containers connected to the data model.
        views: Views connected to the data model.

    """

    space: str
    external_id: str
    version: str
    description: str | None = None
    name: str | None = None
    containers: dict[str, ContainerApply]
    views: dict[str, ViewApply]

    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )

    @classmethod
    def from_rules(
        cls, rules: Rules, data_model_id: dm.DataModelId | None = None, set_expected_source: bool = True
    ) -> Self:
        """Generates a DataModel class instance from a Rules instance.

        Args:
            rules: instance of Rules.
            data_model_id: The id of the data model to be created.
            set_expected_source: Whether to set the expected source on views with direct relation properties.

        Returns:
            Instance of DataModel.
        """
        space = (data_model_id and data_model_id.space) or rules.metadata.cdf_space_name
        external_id = (data_model_id and data_model_id.external_id) or rules.metadata.data_model_name or "missing"
        version = (data_model_id and data_model_id.version) or rules.metadata.version
        names_compliant, name_warnings = are_entity_names_dms_compliant(rules, return_report=True)
        if not names_compliant:
            logging.error(
                exceptions.EntitiesContainNonDMSCompliantCharacters(
                    report=generate_exception_report(name_warnings)
                ).message
            )
            raise exceptions.EntitiesContainNonDMSCompliantCharacters(report=generate_exception_report(name_warnings))

        properties_redefined, redefinition_warnings = are_properties_redefined(rules, return_report=True)
        if properties_redefined:
            logging.error(
                exceptions.PropertiesDefinedMultipleTimes(report=generate_exception_report(redefinition_warnings))
            )
            raise exceptions.PropertiesDefinedMultipleTimes(report=generate_exception_report(redefinition_warnings))

        if rules.metadata.data_model_name is None:
            logging.error(exceptions.DataModelNameMissing(prefix=rules.metadata.prefix).message)
            raise exceptions.DataModelNameMissing(prefix=rules.metadata.prefix)

        return cls(
            space=space,
            external_id=external_id,
            version=version,
            description=rules.metadata.description,
            name=rules.metadata.title,
            containers=cls.containers_from_rules(rules, space),
            views=cls.views_from_rules(rules, space),
        )

    @staticmethod
    def containers_from_rules(rule: Rules, space: str | None = None) -> dict[str, ContainerApply]:
        """Create a dictionary of ContainerApply instances from a Rules instance.

        Args:
            rule: instance of Rules.`
            space: Name of the space to place the containers.

        Returns:
            Dictionary of ContainerApply instances.
        """
        class_properties = to_class_property_pairs(rule)
        return {
            class_id: ContainerApply(
                space=space or rule.metadata.cdf_space_name,
                external_id=class_id,
                name=rule.classes[class_id].class_name,
                description=rule.classes[class_id].description,
                properties=DataModel.container_properties_from_dict(properties),
            )
            for class_id, properties in class_properties.items()
        }

    @staticmethod
    def container_properties_from_dict(properties: dict[str, Property]) -> dict[str, ContainerProperty]:
        """Generates a dictionary of ContainerProperty instances from a dictionary of Property instances.

        Args:
            properties: Dictionary of Property instances.

        Returns:
            Dictionary of ContainerProperty instances.
        """
        container_properties = {}
        for property_id, property_definition in properties.items():
            is_one_to_many = property_definition.max_count != 1
            # Literal, i.e. attribute
            if property_definition.property_type == "DatatypeProperty":
                property_type = cast(
                    type[ListablePropertyType], DATA_TYPE_MAPPING[property_definition.expected_value_type]["dms"]
                )
                container_properties[property_id] = ContainerProperty(
                    type=property_type(is_list=is_one_to_many),
                    nullable=property_definition.min_count == 0,
                    default_value=property_definition.default,
                    name=property_definition.property_name,
                    description=property_definition.description,
                )
            elif property_definition.property_type == "ObjectProperty" and is_one_to_many:
                # One to many edges are only in views and not containers.
                continue

            # URIRef, i.e. edge
            elif property_definition.property_type == "ObjectProperty":
                container_properties[property_id] = ContainerProperty(
                    type=DirectRelation(),
                    nullable=True,
                    name=property_definition.property_name,
                    description=property_definition.description,
                )

            else:
                logging.warning(
                    exceptions.ContainerPropertyTypeUnsupported(property_id, property_definition.property_type).message
                )
                warnings.warn(
                    exceptions.ContainerPropertyTypeUnsupported(property_id, property_definition.property_type).message,
                    category=exceptions.ContainerPropertyTypeUnsupported,
                    stacklevel=2,
                )

        return container_properties

    @staticmethod
    def views_from_rules(rule: Rules, space: str | None = None) -> dict[str, ViewApply]:
        """Generates a dictionary of ViewApply instances from a Rules instance.

        Args:
            rule: Instance of Rules.
            space: Name of the space to place the views.

        Returns:
            Dictionary of ViewApply instances.
        """
        class_properties = to_class_property_pairs(rule)
        space = space or rule.metadata.cdf_space_name
        return {
            class_id: ViewApply(
                space=space,
                external_id=class_id,
                name=rule.classes[class_id].class_name,
                description=rule.classes[class_id].description,
                properties=DataModel.view_properties_from_dict(properties, space, rule.metadata.version),
                version=rule.metadata.version,
            )
            for class_id, properties in class_properties.items()
        }

    @staticmethod
    def view_properties_from_dict(
        properties: dict[str, Property], space: str, version: str
    ) -> dict[str, MappedPropertyApply | ConnectionDefinitionApply]:
        view_properties: dict[str, MappedPropertyApply | ConnectionDefinitionApply] = {}
        for property_id, property_definition in properties.items():
            # attribute
            if property_definition.property_type == "DatatypeProperty":
                view_properties[property_id] = MappedPropertyApply(
                    container=ContainerId(space=space, external_id=property_definition.class_id),
                    container_property_identifier=property_id,
                    name=property_definition.property_name,
                    description=property_definition.description,
                )

            # edge 1-1
            elif property_definition.property_type == "ObjectProperty" and property_definition.max_count == 1:
                view_properties[property_id] = MappedPropertyApply(
                    container=ContainerId(space=space, external_id=property_definition.class_id),
                    container_property_identifier=property_id,
                    name=property_definition.property_name,
                    description=property_definition.description,
                    source=ViewId(space=space, external_id=property_definition.expected_value_type, version=version),
                )

            # edge 1-many
            elif property_definition.property_type == "ObjectProperty" and property_definition.max_count != 1:
                view_properties[property_id] = SingleHopConnectionDefinitionApply(
                    type=DirectRelationReference(
                        space=space, external_id=f"{property_definition.class_id}.{property_definition.property_id}"
                    ),
                    source=ViewId(space=space, external_id=property_definition.expected_value_type, version=version),
                    direction="outwards",
                    name=property_definition.property_name,
                    description=property_definition.description,
                )
            else:
                logging.warning(exceptions.ViewPropertyTypeUnsupported(property_id).message)
                warnings.warn(
                    exceptions.ViewPropertyTypeUnsupported(property_id).message,
                    category=exceptions.ViewPropertyTypeUnsupported,
                    stacklevel=2,
                )

        return view_properties

    def to_cdf(self, client: CogniteClient):
        """Write the the data model to CDF.

        Args:
            client: Connected Cognite client.

        """
        existing_data_model = self.find_existing_data_model(client)
        existing_containers = self.find_existing_containers(client)
        existing_views = self.find_existing_views(client)

        if existing_data_model or existing_containers or existing_views:
            raise exceptions.DataModelOrItsComponentsAlreadyExist(
                existing_data_model, existing_containers, existing_views
            )

        self.create_space(client)
        self.create_containers(client)
        self.create_views(client)
        self.create_data_model(client)

    def find_existing_data_model(self, client: CogniteClient) -> DataModelId | None:
        """Checks if the data model exists in CDF.

        Args:
            client: Cognite client.

        Returns:
             True if the data model exists, False otherwise.
        """
        if model := client.data_modeling.data_models.retrieve((self.space, self.external_id, self.version)):
            cdf_data_model = model.latest_version()
            logging.warning(exceptions.DataModelAlreadyExist(self.external_id, self.version, self.space).message)
            warnings.warn(
                exceptions.DataModelAlreadyExist(self.external_id, self.version, self.space).message,
                category=exceptions.DataModelAlreadyExist,
                stacklevel=2,
            )

            return cdf_data_model.as_id()
        return None

    def find_existing_containers(self, client: CogniteClient) -> set[ContainerId]:
        """Checks if the containers exist in CDF.

        Args:
            client: Cognite client.

        Returns:
            True if the containers exist, False otherwise.
        """

        cdf_containers = {}
        if containers := client.data_modeling.containers.list(space=self.space, limit=-1):
            cdf_containers = {container.as_id(): container for container in containers}

        if existing_containers := {c.as_id() for c in self.containers.values()}.intersection(
            set(cdf_containers.keys())
        ):
            logging.warning(exceptions.ContainersAlreadyExist(existing_containers, self.space).message)
            warnings.warn(
                exceptions.ContainersAlreadyExist(existing_containers, self.space).message,
                category=exceptions.ContainersAlreadyExist,
                stacklevel=2,
            )

            return existing_containers
        else:
            return set()

    def find_existing_views(self, client: CogniteClient) -> set[ViewId]:
        """Checks if the views exist in CDF.

        Args:
            client: Cognite client.

        Returns:
            True if the views exist, False otherwise.
        """
        cdf_views = {}
        if views := client.data_modeling.views.list(space=self.space, limit=-1):
            cdf_views = {view.as_id(): view for view in views if view.version == self.version}

        if existing_views := {v.as_id() for v in self.views.values()}.intersection(set(cdf_views.keys())):
            logging.warning(exceptions.ViewsAlreadyExist(existing_views, self.version, self.space).message)
            warnings.warn(
                exceptions.ViewsAlreadyExist(existing_views, self.version, self.space).message,
                category=exceptions.ViewsAlreadyExist,
                stacklevel=2,
            )
            return existing_views
        else:
            return set()

    def create_space(self, client: CogniteClient):
        logging.info(f"Creating space {self.space}")
        _ = client.data_modeling.spaces.apply(SpaceApply(space=self.space))

    def create_containers(self, client: CogniteClient):
        for container_id in self.containers:
            logging.info(f"Creating container {container_id} in space {self.space}")
            _ = client.data_modeling.containers.apply(self.containers[container_id])

    def create_views(self, client: CogniteClient):
        for view_id in self.views:
            logging.info(f"Creating view {view_id} version {self.views[view_id].version} in space {self.space}")
            _ = client.data_modeling.views.apply(self.views[view_id])

    def create_data_model(self, client: CogniteClient):
        logging.info(f"Creating data model {self.external_id} version {self.version} in space {self.space}")
        _ = client.data_modeling.data_models.apply(
            DataModelApply(
                name=self.name,
                description=self.description,
                space=self.space,
                external_id=self.external_id,
                version=self.version,
                views=list(self.views.values()),
            )
        )

    def remove_data_model(self, client: CogniteClient):
        """Helper function to remove a data model, and all underlying views and containers from CDF.

        Args:
            client: Cognite client.
        """

        if client.data_modeling.data_models.retrieve((self.space, self.external_id, self.version)):
            logging.info(f"Removing data model {self.external_id} version {self.version} from space {self.space}")
            _ = client.data_modeling.data_models.delete((self.space, self.external_id, self.version))

        if views := client.data_modeling.views.retrieve(
            [view.as_id() for view in self.views.values()], all_versions=False
        ):
            for view in views:
                logging.info(f"Removing view {view.external_id} version {view.version} from space {self.space}")
                _ = client.data_modeling.views.delete((view.space, view.external_id, view.version))

        if containers := client.data_modeling.containers.retrieve(
            [container.as_id() for container in self.containers.values()]
        ):
            for container in containers:
                logging.info(f"Removing container {container.external_id} from space {self.space}")
                _ = client.data_modeling.containers.delete((container.space, container.external_id))



from abc import ABC, abstractmethod
from pathlib import Path
from typing import ClassVar, TypeVar

from pydantic import BaseModel, ConfigDict

from cognite.neat.app.monitoring.metrics import NeatMetricsCollector
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigs


class Config(BaseModel):
    ...


class Configurable(BaseModel):
    name: str
    value: str | None = None
    label: str | None = None
    type: str | None = None  # string , secret , number , boolean , json
    required: bool = False
    options: list[str] | None = None


class DataContract(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)


T_Input = TypeVar("T_Input", bound=DataContract)
T_Input1 = TypeVar("T_Input1", bound=DataContract)
T_Input2 = TypeVar("T_Input2", bound=DataContract)
T_Output = TypeVar("T_Output", bound=DataContract)


class Step(ABC):
    description: str = ""
    category: str = "default"
    configurables: ClassVar[list[Configurable]] = []
    scope: str = "core_global"
    metrics: NeatMetricsCollector | None = None
    workflow_configs: WorkflowConfigs | None = None
    version: str = "1.0.0"  # version of the step. All alpha versions considered as experimental
    source: str = (
        "cognite"  # source of the step , can be source identifier or url , for instance github url for instance.
    )
    docs_url: str = "https://cognite-neat.readthedocs-hosted.com/en/latest/"  # url to the documentation of the step

    def __init__(self, data_store_path: Path | None = None):
        self.log: bool = False
        self.configs: dict[str, str] = {}
        self.workflow_id: str = ""
        self.workflow_run_id: str = ""
        self.data_store_path = Path(data_store_path) if data_store_path is not None else Path.cwd()

    @property
    def _not_configured_message(self) -> str:
        return f"Step {type(self).__name__} has not been configured."

    def set_metrics(self, metrics: NeatMetricsCollector | None):
        self.metrics = metrics

    def set_workflow_configs(self, configs: WorkflowConfigs | None):
        self.workflow_configs = configs

    def set_workflow_metadata(self, workflow_id: str, workflow_run_id: str):
        self.workflow_id = workflow_id
        self.workflow_run_id = workflow_run_id

    def configure(self, configs: dict[str, str]):
        self.configs = configs

    def set_flow_context(self, context: dict[str, DataContract]):
        self.flow_context = context

    @abstractmethod
    def run(self, *input_data: DataContract) -> DataContract | tuple[FlowMessage, DataContract] | FlowMessage:
        ...



import json
import logging
import time
from datetime import datetime
from pathlib import Path
from typing import ClassVar

import yaml
from rdflib import Namespace

from cognite.neat.rules import exporter, importer
from cognite.neat.rules.models.rdfpath import TransformationRuleType
from cognite.neat.rules.models.rules import Class, Classes, Metadata, Properties, Property, Rules
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.steps.data_contracts import RulesData, SolutionGraph, SourceGraph
from cognite.neat.workflows.steps.step_model import Configurable, Step

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()

__all__ = ["OpenApiToRules", "ArbitraryJsonYamlToRules", "GraphToRules", "OntologyToRules"]


class OntologyToRules(Step):
    """The step extracts schema from OpenApi/Swagger specification and generates NEAT transformation rules object."""

    description = "The step extracts NEAT rules object from OWL Ontology and \
        exports them as an Excel rules files for further editing."
    category = CATEGORY
    version = "0.1.0-alpha"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="ontology_file_path", value="staging/ontology.ttl", label="Relative path to the OWL ontology file."
        ),
        Configurable(
            name="excel_file_path", value="staging/rules.xlsx", label="Relative path for the Excel rules storage."
        ),
    ]

    def run(self) -> FlowMessage:  # type: ignore[override, syntax]
        ontology_file_path = self.data_store_path / Path(self.configs["ontology_file_path"])
        excel_file_path = self.data_store_path / Path(self.configs["excel_file_path"])
        report_file_path = excel_file_path.parent / f"report_{excel_file_path.stem}.txt"

        rules = importer.OWLImporter(ontology_file_path).to_rules(skip_validation=True)
        assert isinstance(rules, Rules)
        exporter.ExcelExporter.from_rules(rules).export_to_file(excel_file_path)

        if report := importer.ExcelImporter(filepath=excel_file_path).to_raw_rules().validate_rules():
            report_file_path.write_text(report)

        relative_excel_file_path = str(excel_file_path).split("/data/")[1]
        relative_report_file_path = str(report_file_path).split("/data/")[1]

        output_text = (
            "<p></p>"
            "Rules generated from OWL Ontology can be downloaded here : "
            f'<a href="http://localhost:8000/data/{relative_excel_file_path}?{time.time()}" '
            f'target="_blank">{excel_file_path.stem}.xlsx</a>'
            "<p></p>"
            "Report can be downloaded here : "
            f'<a href="http://localhost:8000/data/{relative_report_file_path}?{time.time()}" '
            f'target="_blank">{report_file_path.stem}.txt</a>'
        )

        return FlowMessage(output_text=output_text)


class OpenApiToRules(Step):
    """The step extracts schema from OpenApi/Swagger specification and generates NEAT transformation rules object."""

    description = "The step extracts schema from OpenAPI specification and generates NEAT transformation rules object. \
    The rules object can be serialized to excel file or used directly in other steps."
    category = CATEGORY
    version = "0.1.0-alpha"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="openapi_spec_file_path",
            value="workflows/openapi_to_rules/source_data/openapi.json",
            label="Relative path to the OpenAPI spec file.The file can be in either json or in yaml format.",
        ),
        Configurable(
            name="fdm_compatibility_mode",
            value="True",
            label="If set to True, the step will try to convert property names to FDM compatible names.",
            options=["True", "False"],
        ),
    ]

    def run(self) -> (FlowMessage, RulesData):  # type: ignore[override, syntax]
        openapi_file_path = self.data_store_path / Path(self.configs["openapi_spec_file_path"])
        self.processed_classes_counter = 0
        self.processed_properies_counter = 0
        self.failed_classes_counter = 0
        self.failed_properties_counter = 0
        self.failed_classes: dict[str, str] = {}
        self.failed_properties: dict[str, str] = {}
        self.is_fdm_compatibility_mode = self.configs["fdm_compatibility_mode"] == "True"
        rules = self.open_api_to_rules(openapi_file_path)
        report = f" Generated Rules and source data model from OpenApi specs   \
        <p> Processed {self.processed_classes_counter} classes and {self.processed_properies_counter} properties. </p> \
        <p> Failed to process {self.failed_classes_counter} classes and \
        {self.failed_properties_counter} properties. </p>"
        report_obj = {
            "processed_classes_counter": self.processed_classes_counter,
            "processed_properies_counter": self.processed_properies_counter,
            "failed_classes": self.failed_classes,
            "failed_properties": self.failed_properties,
        }
        return (FlowMessage(output_text=report, payload=report_obj), RulesData(rules=rules))

    def open_api_to_rules(self, open_api_spec_file_path: Path) -> Rules:
        """Converts OpenAPI spec to NEAT transformation rules object."""
        with open_api_spec_file_path.open("r") as openapi_file:
            if open_api_spec_file_path.suffix == ".json":
                openapi_spec = json.load(openapi_file)
            elif open_api_spec_file_path.suffix == ".yaml":
                openapi_spec = yaml.safe_load(openapi_file)

        metadata = Metadata(
            title="OpenAPI to DM transformation rules",
            description="OpenAPI to DM transformation rules",
            version="0.1",
            creator="Cognite",
            created=datetime.utcnow(),
            namespace=Namespace("http://purl.org/cognite/neat#"),
            prefix="neat",
            data_model_name="OpenAPI",
            cdf_space_name="OpenAPI",
        )

        classes = Classes()
        properties = Properties()

        # Loop through OpenAPI components
        for component_name, component_info in openapi_spec.get("components", {}).get("schemas", {}).items():
            if self.is_fdm_compatibility_mode:
                class_name = get_dms_compatible_name(create_fdm_compatibility_class_name(component_name))
            else:
                class_name = component_name
            class_id = class_name
            logging.info(f" OpenAPi parser : Processing class {class_id} ")
            try:
                class_ = Class(
                    class_id=class_id,
                    class_name=class_name,
                    description=component_info.get("description", component_info.get("title", "empty")),
                )
                classes[class_id] = class_
                self.processed_classes_counter += 1
                # Loop through properties of OpenApi spec]
                self.process_properies(properties, class_id, class_name, component_info)

            except Exception as e:
                logging.error(f" OpenAPi parser : Error creating class {class_id}: {e}")
                self.failed_classes_counter += 1
                self.failed_classes[class_id] = str(e)

        rules = Rules(metadata=metadata, classes=classes, properties=properties, prefixes={}, instances=[])

        return rules

    def process_properies(
        self,
        rules_properties: Properties,
        class_id: str,
        class_name: str,
        component: dict,
        parent_property_name: str | None = None,
    ):
        # component can have keys : type, description, properties, required, title, allOf, anyOf, oneOf§
        logging.info(f" OpenAPi parser : Processing properties for class {class_id} , component {component}")
        for component_name, component_info in component.items():
            if component_name == "allOf" or component_name == "anyOf" or component_name == "oneOf":
                for sub_component in component_info:
                    self.process_properies(rules_properties, class_id, class_name, sub_component, component_name)
            elif component_name == "properties":
                for prop_name, prop_info in component_info.items():
                    prop_id = prop_name
                    if prop_name == "allOf" or prop_name == "anyOf" or prop_name == "oneOf":
                        if isinstance(prop_info, list):
                            prop_type = prop_info[0].get("type", "string")
                        else:
                            logging.error(f" !!!!! prop info is not a list . its: {prop_info} ")

                    else:
                        prop_type = prop_info.get("type", "string")
                        if prop_type == "array":
                            self.process_properies(
                                rules_properties, class_id, class_name, prop_info.get("items", {}), prop_name
                            )
                            continue
                    expected_value_type = self.map_open_api_type(prop_type)
                    if prop_name == "$ref":
                        ref_class = self.get_ref_class_name(prop_info.get("$ref", ""))
                        expected_value_type = ref_class
                    try:
                        prop = Property(
                            class_id=class_id,
                            property_id=get_dms_compatible_name(prop_id) if self.is_fdm_compatibility_mode else prop_id,
                            property_name=get_dms_compatible_name(prop_name)
                            if self.is_fdm_compatibility_mode
                            else prop_name,
                            property_type="ObjectProperty",
                            description=prop_info.get("description", prop_info.get("title", "empty")),
                            expected_value_type=expected_value_type,
                            cdf_resource_type=["Asset"],
                            resource_type_property="Asset",  # type: ignore
                            rule_type=TransformationRuleType("rdfpath"),
                            rule=f"neat:{class_name}(neat:{prop_name})",
                            label="linked to",
                        )
                        self.processed_properies_counter += 1
                        rules_properties[class_id + prop_id] = prop
                    except Exception as e:
                        logging.error(f" OpenAPi parser : Error creating property {prop_id}: {e}")
                        self.failed_properties_counter += 1
                        self.failed_properties[prop_id] = str(e)
            elif component_name == "$ref":
                ref_class = self.get_ref_class_name(component_info)
                if parent_property_name is not None:
                    prop = Property(
                        class_id=class_id,
                        property_id=parent_property_name,
                        property_name=parent_property_name,
                        property_type="ObjectProperty",
                        description="no",
                        expected_value_type=ref_class,
                        cdf_resource_type=["Asset"],
                        resource_type_property="Asset",  # type: ignore
                        rule_type=TransformationRuleType("rdfpath"),
                        rule=f"neat:{class_name}(neat:{parent_property_name})",
                        label="linked to",
                    )
                    rules_properties[class_id + parent_property_name] = prop

    def get_ref_class_name(self, ref: str) -> str:
        ref_payload = ref.split("/")[-1]
        if self.is_fdm_compatibility_mode:
            return get_dms_compatible_name(create_fdm_compatibility_class_name(ref_payload))
        return ref_payload

    def map_open_api_type(self, openapi_type: str) -> str:
        """Map OpenAPI type to NEAT compatible types"""
        if openapi_type == "object":
            datatype = "json"
        elif openapi_type == "array":
            datatype = "sequence"
        elif openapi_type == "number":
            datatype = "float"
        else:
            return openapi_type  # Default to string
        return datatype


class ArbitraryJsonYamlToRules(Step):
    """The step extracts schema from arbitrary json or yaml file and generates NEAT transformation rules object."""

    description = "The step extracts schema from arbitrary json file and generates NEAT transformation rules object."
    category = CATEGORY
    version = "0.1.0-alpha"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_path",
            value="workflows/openapi_to_rules/data/data.json",
            label="Relative path to the json file.The file can be in either json or in yaml format.",
        ),
        Configurable(
            name="fdm_compatibility_mode",
            value="True",
            label="If set to True, the step will try to convert property names to FDM compatible names.",
            options=["True", "False"],
        ),
    ]

    def run(self) -> (FlowMessage, RulesData):  # type: ignore[override, syntax]
        openapi_file_path = Path(self.data_store_path) / Path(self.configs["file_path"])
        self.processed_classes_counter = 0
        self.processed_properies_counter = 0
        self.failed_classes_counter = 0
        self.failed_properties_counter = 0
        self.failed_classes: dict[str, str] = {}
        self.failed_properties: dict[str, str] = {}
        self.is_fdm_compatibility_mode = self.configs["fdm_compatibility_mode"] == "True"

        rules = self.dict_to_rules(openapi_file_path)
        report = f" Generated Rules and source data model from json/yaml   \
        <p> Processed {self.processed_classes_counter} classes and {self.processed_properies_counter} properties. </p> \
        <p> Failed to process {self.failed_classes_counter} classes \
        and {self.failed_properties_counter} properties. </p>"
        report_obj = {
            "processed_classes_counter": self.processed_classes_counter,
            "processed_properies_counter": self.processed_properies_counter,
            "failed_classes": self.failed_classes,
            "failed_properties": self.failed_properties,
        }
        return FlowMessage(output_text=report, payload=report_obj), RulesData(rules=rules)

    def dict_to_rules(self, open_api_spec_file_path: Path) -> Rules:
        """Converts OpenAPI spec to NEAT transformation rules object."""
        with open_api_spec_file_path.open("r") as openapi_file:
            if open_api_spec_file_path.suffix == ".json":
                src_data_obj = json.load(openapi_file)
            elif open_api_spec_file_path.suffix == ".yaml":
                src_data_obj = yaml.safe_load(openapi_file)

        metadata = Metadata(
            title="OpenAPI to DM transformation rules",
            description="OpenAPI to DM transformation rules",
            version="0.1",
            creator="Cognite",
            created=datetime.utcnow(),
            namespace=Namespace("http://purl.org/cognite/neat#"),
            prefix="neat",
            cdf_space_name="OpenAPI",
            data_model_name="OpenAPI",
        )

        self.classes = Classes()
        self.properties = Properties()

        self.convert_dict_to_classes_and_props(src_data_obj, None)
        rules = Rules(metadata=metadata, classes=self.classes, properties=self.properties, prefixes={}, instances=[])

        return rules

    def add_class(self, class_name: str, description: str | None = None, parent_class_name: str | None = None):
        if class_name in self.classes:
            return
        if self.is_fdm_compatibility_mode:
            class_name = get_dms_compatible_name(create_fdm_compatibility_class_name(class_name))
        try:
            class_ = Class(class_id=class_name, class_name=class_name, description=description)
            if parent_class_name:
                self.add_property(class_name, "parent", parent_class_name, None)
            self.classes[class_name] = class_
            self.processed_classes_counter += 1
        except Exception as e:
            logging.error(f" OpenAPi parser : Error creating class {class_name}: {e}")
            self.failed_classes_counter += 1
            self.failed_classes[class_name] = str(e)
        return

    def add_property(self, class_name: str, property_name: str, property_type: str, description: str | None = None):
        if class_name + property_name in self.properties:
            return
        if self.is_fdm_compatibility_mode:
            property_name = get_dms_compatible_name(property_name)
            class_name = get_dms_compatible_name(create_fdm_compatibility_class_name(class_name))
        try:
            prop = Property(
                class_id=class_name,
                property_id=property_name,
                property_name=property_name,
                property_type="ObjectProperty",
                description=description,
                expected_value_type=property_type,
                cdf_resource_type=["Asset"],
                resource_type_property="Asset",  # type: ignore
                rule_type=TransformationRuleType("rdfpath"),
                rule=f"neat:{class_name}(neat:{property_name})",
                label="linked to",
            )
            self.properties[class_name + property_name] = prop
            self.processed_properies_counter += 1
        except Exception as e:
            logging.error(f" OpenAPi parser : Error creating property {property_name}: {e}")
            self.failed_properties_counter += 1
            self.failed_properties[class_name + property_name] = str(e)
        return

    # Iterate through the JSON data and convert it to triples
    def convert_dict_to_classes_and_props(self, data: dict, parent_property_name=None, grand_parent_property_name=None):
        if isinstance(data, dict):
            if len(data) == 0:
                return
            if parent_property_name is None:
                for key, value in data.items():
                    self.convert_dict_to_classes_and_props(value, key)
            else:
                description = None
                self.add_class(parent_property_name, description, grand_parent_property_name)
                for key, value in data.items():
                    self.convert_dict_to_classes_and_props(value, key, parent_property_name)
        elif isinstance(data, list):
            for item in data:
                self.convert_dict_to_classes_and_props(item, parent_property_name, grand_parent_property_name)
        else:
            # Convert scalar values to RDF literals
            data_type = ""
            if isinstance(data, bool):
                data_type = "boolean"
            elif isinstance(data, int):
                data_type = "integer"
            elif isinstance(data, float):
                data_type = "float"
            elif isinstance(data, str):
                data_type = "string"
            else:
                data_type = "string"
            if grand_parent_property_name is None:
                logging.error(" grand_parent_property_name is None")
                return

            self.add_property(grand_parent_property_name, parent_property_name, data_type, None)


def get_dms_compatible_name(name: str) -> str:
    """Converts name to DMS compatible name.It applies both to class and property names"""
    # reserverd words in DMS
    reserved_words_mapping = {
        "space": "src_space",
        "externalId": "external_id",
        "createdTime": "created_time",
        "lastUpdatedTime": "last_updated_time",
        "deletedTime": "deleted_time",
        "edge_id": "src_edge_id",
        "node_id": "src_node_id",
        "project_id": "src_project_id",
        "property_group": "src_property_group",
        "seq": "src_seq",
        "tg_table_name": "src__table_name",
        "extensions": "src_extensions",
    }
    if name in reserved_words_mapping:
        return reserved_words_mapping[name]
    else:
        return name.replace(".", "_")


def create_fdm_compatibility_class_name(input_string: str):
    """Remove underscores and capitalize each word in the string ,
    the conversion is done to improve compliance with DMS naming conventions"""

    if "_" in input_string:
        words = input_string.split("_")  # Split the string by underscores
        result = "".join([word.capitalize() for word in words])  # Capitalize each word
        return result
    else:
        return input_string


class GraphToRules(Step):
    """The step extracts data model from RDF graph and generates NEAT transformation rules object."""

    description = "The step extracts data model from RDF graph and generates NEAT transformation rules object. \
    The rules object can be serialized to excel file or used directly in other steps."
    category = CATEGORY
    version = "0.1.0-alpha"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_name", value="inferred_transformations.xlsx", label="File name to store transformation rules."
        ),
        Configurable(name="storage_dir", value="staging", label="Directory to store Transformation Rules spreadsheet"),
        Configurable(
            name="max_number_of_instances",
            value="-1",
            label="Maximum number of instances per class to process, -1 means all instances",
        ),
    ]

    def run(self, graph_store: SourceGraph | SolutionGraph) -> FlowMessage:  # type: ignore[override, syntax]
        file_name = self.configs["file_name"]
        staging_dir_str = self.configs["storage_dir"]
        staging_dir = self.data_store_path / Path(staging_dir_str)
        staging_dir.mkdir(parents=True, exist_ok=True)

        spreadsheet_path = staging_dir / file_name

        raw_rules = importer.GraphImporter(
            graph_store.graph.graph, int(self.configs["max_number_of_instances"])
        ).to_raw_rules()
        exporter.ExcelExporter.from_rules(raw_rules).export_to_file(spreadsheet_path)

        output_text = (
            "<p></p>"
            "Rules imported using GraphImporter and can be downloaded as the Excel file:"
            f'<a href="http://localhost:8000/data/{staging_dir_str}/{file_name}?{time.time()}" '
            f'target="_blank">{file_name}</a>'
            "<p></p>"
            "Rules validation report can be downloaded as TXT file: "
            f'<a href="http://localhost:8000/data/{staging_dir_str}/{spreadsheet_path.stem}'
            f'_validation_report.txt?{time.time()}" '
            f'target="_blank">{spreadsheet_path.stem}_validation_report.txt</a>'
        )

        return FlowMessage(output_text=output_text)



from .configuration import RdfStoreConfig, RdfStoreType
from .graph_store import NeatGraphStore, drop_graph_store_storage

__all__ = ["NeatGraphStore", "drop_graph_store_storage", "RdfStoreType", "RdfStoreConfig"]



from __future__ import annotations

from datetime import date, datetime
from typing import cast

from cognite.client.data_classes.data_modeling import (
    Boolean,
    Date,
    FileReference,
    Float64,
    Int32,
    Int64,
    Json,
    PropertyType,
    SequenceReference,
    Text,
    TimeSeriesReference,
    Timestamp,
)

_DATA_TYPES: list[dict[str, str | type]] = [
    {"name": "boolean", "python": bool, "GraphQL": "Boolean", "dms": Boolean},
    {"name": "float", "python": float, "GraphQL": "Float", "dms": Float64},
    {"name": "integer", "python": int, "GraphQL": "Int", "dms": Int32},
    {"name": "nonPositiveInteger", "python": int, "GraphQL": "Int", "dms": Int32},
    {"name": "nonNegativeInteger", "python": int, "GraphQL": "Int", "dms": Int32},
    {"name": "negativeInteger", "python": "int", "GraphQL": "Int", "dms": Int32},
    {"name": "long", "python": int, "GraphQL": "Int", "dms": Int64},
    {"name": "string", "python": str, "GraphQL": "String", "dms": Text},
    {"name": "anyURI", "python": str, "GraphQL": "String", "dms": Text},
    {"name": "normalizedString", "python": str, "GraphQL": "String", "dms": Text},
    {"name": "token", "python": str, "GraphQL": "String", "dms": Text},
    # Graphql does not have a datetime/date type this is CDF specific
    {"name": "dateTime", "python": datetime, "GraphQL": "Timestamp", "dms": Timestamp},
    {"name": "date", "python": date, "GraphQL": "String", "dms": Date},
    # CDF specific types, not in XSD
    {"name": "timeseries", "python": TimeSeriesReference, "GraphQL": "TimeSeries", "dms": TimeSeriesReference},
    {"name": "file", "python": FileReference, "GraphQL": "File", "dms": FileReference},
    {"name": "sequence", "python": SequenceReference, "GraphQL": "Sequence", "dms": TimeSeriesReference},
    {"name": "json", "python": Json, "GraphQL": "Json", "dms": Json},
]


# mapping of XSD types to Python and GraphQL types
DATA_TYPE_MAPPING: dict[str, dict[str, type | str]] = {
    cast(str, entry["name"]): {key: entry[key] for key in ["python", "GraphQL", "dms"]} for entry in _DATA_TYPES
}


DMS_TO_DATA_TYPE: dict[type[PropertyType], str] = {}
for entry in _DATA_TYPES:
    if entry["dms"] not in DMS_TO_DATA_TYPE:
        DMS_TO_DATA_TYPE[cast(type[PropertyType], entry["dms"])] = cast(str, entry["name"])


def type_to_target_convention(type_: str, target_type_convention: str) -> type | str | PropertyType:
    """Returns the GraphQL type for a given XSD type."""
    return DATA_TYPE_MAPPING[type_][target_type_convention]



import logging
import time
import warnings
from pathlib import Path
from typing import ClassVar

import cognite.neat.graph.extractors.graph_sheet_to_graph
from cognite.neat.exceptions import wrangle_warnings
from cognite.neat.rules import exporter
from cognite.neat.rules.exporter._rules2dms import DataModel
from cognite.neat.rules.exporter._rules2graphql import GraphQLSchema
from cognite.neat.rules.exporter._rules2ontology import Ontology
from cognite.neat.utils.utils import generate_exception_report
from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.steps.data_contracts import CogniteClient, DMSDataModel, RulesData
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = [
    "DMSDataModelFromRules",
    "GraphQLSchemaFromRules",
    "OntologyFromRules",
    "SHACLFromRules",
    "GraphCaptureSpreadsheetFromRules",
    "UploadDMSDataModel",
    "DeleteDMSDataModel",
    "ExcelFromRules",
]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()


class DMSDataModelFromRules(Step):
    """
    This step generates DMS Data model from data model defined in transformation rules
    """

    description = "This step generates DMS Data model from data model defined in transformation rules."
    category = CATEGORY

    def run(self, transformation_rules: RulesData) -> (FlowMessage, DMSDataModel):  # type: ignore[override, syntax]
        data_model = DataModel.from_rules(transformation_rules.rules)

        output_text = (
            f"DMS Data Model <b><code>{data_model.external_id}</code></b> version"
            f" <b><code>{data_model.version}</code></b> generated containing:<ul>"
            f"<li> {len(data_model.containers)} containers</li>"
            f"<li> {len(data_model.views)} views</li>"
            f"</ul> Data model is meant to be uploaded to <b><code>{data_model.space}</code></b> space"
        )

        return FlowMessage(output_text=output_text), DMSDataModel(data_model=data_model)


class UploadDMSDataModel(Step):
    """
    This step uploaded generated DMS Data model
    """

    description = "This step uploaded generated DMS Data model."
    category = CATEGORY

    def run(self, data_model: DMSDataModel, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override, syntax]
        data_model.data_model.to_cdf(cdf_client)

        output_text = (
            f"DMS Data Model <b><code>{data_model.data_model.external_id}</code></b> version"
            f" <b><code>{data_model.data_model.version}</code></b> uploaded to space"
            f" <b><code>{data_model.data_model.space}</code></b> containing:<ul>"
            f"<li> {len(data_model.data_model.containers)} containers</li>"
            f"<li> {len(data_model.data_model.views)} views</li></ul>"
        )

        return FlowMessage(output_text=output_text)


class DeleteDMSDataModel(Step):
    """
    This step deletes DMS Data model and all underlying containers and views
    """

    description = "This step deletes DMS Data model and all underlying containers and views."
    category = CATEGORY

    def run(self, data_model: DMSDataModel, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override, syntax]
        data_model.data_model.remove_data_model(cdf_client)

        output_text = (
            f"DMS Data Model {data_model.data_model.external_id} version {data_model.data_model.version} "
            f"under {data_model.data_model.space} removed:"
            f"<p> - {len(data_model.data_model.containers)} containers removed</p>"
            f"<p> - {len(data_model.data_model.views)} views removed</p>"
        )

        output_text = (
            f"DMS Data Model <b><code>{data_model.data_model.external_id}</code></b> version"
            f" <b><code>{data_model.data_model.version}</code></b> removed"
            f" from space <b><code>{data_model.data_model.space}</code></b> as well:"
            f"<ul><li> {len(data_model.data_model.containers)} containers</li>"
            f"<li> {len(data_model.data_model.views)} views</li></ul>"
        )

        return FlowMessage(output_text=output_text)


class GraphQLSchemaFromRules(Step):
    """
    This step generates GraphQL schema from data model defined in transformation rules
    """

    description = "This step generates GraphQL schema from data model defined in transformation rules."
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_name",
            value="",
            label=(
                "Name of the GraphQL schema file it must have .graphql extension,"
                " if empty defaults to form `prefix-version.graphql`"
            ),
        ),
        Configurable(name="storage_dir", value="staging", label="Directory to store GraphQL schema file"),
    ]

    def run(self, transformation_rules: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        data_model_gql = GraphQLSchema.from_rules(transformation_rules.rules, verbose=True).schema

        default_name = (
            f"{transformation_rules.rules.metadata.prefix}-"
            f"v{transformation_rules.rules.metadata.version.strip().replace('.', '_')}"
            ".graphql"
        )

        schema_name = self.configs["file_name"] or default_name

        staging_dir_str = self.configs["storage_dir"]
        staging_dir = self.data_store_path / Path(staging_dir_str)
        staging_dir.mkdir(parents=True, exist_ok=True)
        fdm_model_full_path = staging_dir / schema_name

        fdm_model_full_path.write_text(data_model_gql)

        output_text = (
            "<p></p>"
            "GraphQL Schema generated and can be downloaded here : "
            f'<a href="http://localhost:8000/data/{staging_dir_str}/{schema_name}?{time.time()}" '
            f'target="_blank">{schema_name}</a>'
        )

        return FlowMessage(output_text=output_text)


class OntologyFromRules(Step):
    """
    This step generates OWL ontology from data model defined in transformation rules
    """

    description = "This step generates OWL ontology from data model defined in transformation rules."
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_name",
            value="",
            label=(
                "Name of the OWL ontology file it must have .ttl extension,"
                " if empty defaults to form `prefix-version-ontology.ttl`"
            ),
        ),
        Configurable(name="storage_dir", value="staging", label="Directory to store the OWL ontology file"),
        Configurable(
            name="store_warnings",
            value="True",
            label="To store warnings while generating ontology",
            options=["True", "False"],
        ),
    ]

    def run(self, transformation_rules: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        # ontology file
        default_name = (
            f"{transformation_rules.rules.metadata.prefix}-"
            f"v{transformation_rules.rules.metadata.version.strip().replace('.', '_')}"
            "-ontology.ttl"
        )

        ontology_file = self.configs["file_name"] or default_name

        storage_dir_str = self.configs["storage_dir"]
        storage_dir = self.data_store_path / storage_dir_str
        storage_dir.mkdir(parents=True, exist_ok=True)

        store_warnings = self.configs["store_warnings"].lower() == "true"

        with warnings.catch_warnings(record=True) as validation_warnings:
            ontology = Ontology.from_rules(transformation_rules=transformation_rules.rules)

        with (storage_dir / ontology_file).open(mode="w") as onto_file:
            onto_file.write(ontology.ontology)

        if store_warnings and validation_warnings:
            with (storage_dir / "report.txt").open(mode="w") as report_file:
                report_file.write(generate_exception_report(wrangle_warnings(validation_warnings), "Warnings"))

        output_text = (
            "<p></p>"
            "Ontology generated and can be downloaded here : "
            f'<a href="http://localhost:8000/data/{storage_dir_str}/{ontology_file}?{time.time()}" '
            f'target="_blank">{ontology_file}</a>'
        )

        output_text += (
            (
                "<p></p>"
                " Download conversion report "
                f'<a href="http://localhost:8000/data/{storage_dir_str}/report.txt?{time.time()}" '
                f'target="_blank">here</a>'
            )
            if validation_warnings
            else ""
        )

        return FlowMessage(output_text=output_text)


class SHACLFromRules(Step):
    """
    This step generates SHACL from data model defined in transformation rules
    """

    description = "This step generates SHACL from data model defined in transformation rules"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_name",
            value="",
            label=(
                "Name of the SHACL file it must have .ttl extension, if "
                "empty defaults to form `prefix-version-shacl.ttl`"
            ),
        ),
        Configurable(name="storage_dir", value="staging", label="Directory to store the SHACL file"),
    ]

    def run(self, transformation_rules: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        # ontology file
        default_name = (
            f"{transformation_rules.rules.metadata.prefix}-"
            f"v{transformation_rules.rules.metadata.version.strip().replace('.', '_')}"
            "-shacl.ttl"
        )

        shacl_file = self.configs["file_name"] or default_name

        storage_dir_str = self.configs["storage_dir"]
        storage_dir = self.data_store_path / storage_dir_str
        storage_dir.mkdir(parents=True, exist_ok=True)

        constraints = Ontology.from_rules(transformation_rules=transformation_rules.rules).constraints

        with (storage_dir / shacl_file).open(mode="w") as onto_file:
            onto_file.write(constraints)

        output_text = (
            "<p></p>"
            "SHACL generated and can be downloaded here : "
            f'<a href="http://localhost:8000/data/{storage_dir_str}/{shacl_file}?{time.time()}" '
            f'target="_blank">{shacl_file}</a>'
        )
        return FlowMessage(output_text=output_text)


class GraphCaptureSpreadsheetFromRules(Step):
    """
    This step generates data capture spreadsheet from data model defined in rules
    """

    description = "This step generates data capture spreadsheet from data model defined in rules"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="file_name", value="graph_capture_sheet.xlsx", label="File name of the data capture sheet"),
        Configurable(name="auto_identifier_type", value="index-based", label="Type of automatic identifier"),
        Configurable(name="storage_dir", value="staging", label="Directory to store data capture sheets"),
    ]

    def run(self, rules: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        logging.info("Generate graph capture sheet")
        sheet_name = self.configs["file_name"]
        auto_identifier_type = self.configs["auto_identifier_type"]
        staging_dir_str = self.configs["storage_dir"]
        logging.info(f"Auto identifier type {auto_identifier_type}")
        staging_dir = self.data_store_path / Path(staging_dir_str)
        staging_dir.mkdir(parents=True, exist_ok=True)
        data_capture_sheet_path = staging_dir / sheet_name

        cognite.neat.graph.extractors.graph_sheet_to_graph.rules2graph_capturing_sheet(
            rules.rules, data_capture_sheet_path, auto_identifier_type=auto_identifier_type
        )

        output_text = (
            "Data capture sheet generated and can be downloaded here : "
            f'<a href="http://localhost:8000/data/{staging_dir_str}/{sheet_name}?{time.time()}" target="_blank">'
            f"{sheet_name}</a>"
        )
        return FlowMessage(output_text=output_text)


class ExcelFromRules(Step):
    description = "This step generates Excel file from rules"
    category = CATEGORY
    version = "0.1.0-alpha"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="output_file_path", value="rules/custom-rules.xlsx", label="File path to the generated Excel file"
        )
    ]

    def run(self, rules_data: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        full_path = Path(self.data_store_path) / Path(self.configs["output_file_path"])
        exporter.ExcelExporter.from_rules(rules=rules_data.rules).export_to_file(filepath=full_path)
        return FlowMessage(output_text="Generated Excel file from rules")



# import datetime
# import json
# from pathlib import Path

# from rdflib import Namespace
# from cognite.neat.rules.models import Class, Metadata, Property, TransformationRules
# from cognite.neat.workflows.model import FlowMessage
# from cognite.neat.workflows.steps.data_contracts import RulesData

# from cognite.neat.workflows.steps.step_model import Configurable, Step
# import logging
# import yaml


# class OpenApiToRules(Step):
#     """The step extracts schema from OpenAPI spec and generates NEAT transformation rules object."""
#     description = "The step extracts schema from OpenAPI spec and generates NEAT transformation rules object."
#     category = "openapi"
#     version = "0.1.0-alpha"
#     configurables = [
#         Configurable(
#             name="openapi_spec_file_path",
#             value="workflows/openapi_to_rules/source_data/openapi.json",
#             label="Relative path to the OpenAPI spec file.The file can be in either json or yaml format.",
#         )
#     ]

#     def run(self) -> (FlowMessage, RulesData):
#         openapi_file_path = Path(self.data_store_path) / Path(self.configs["openapi_spec_file_path"])
#         rules = self.open_api_to_rules(openapi_file_path)
#         return (FlowMessage(output_text="Generated Rules from OpenApi spec"), RulesData(rules=rules))

#     def open_api_to_rules(self, open_api_spec_file_path: Path) -> TransformationRules:
#         """Converts OpenAPI spec to NEAT transformation rules object. """
#         with open_api_spec_file_path.open("r") as openapi_file:
#             if open_api_spec_file_path.suffix == ".json":
#                 openapi_spec = json.load(openapi_file)
#             elif open_api_spec_file_path.suffix == ".yaml":
#                 openapi_spec = yaml.safe_load(openapi_file)

#         metadata = Metadata(
#             title="OpenAPI to DM transformation rules",
#             description="OpenAPI to DM transformation rules",
#             version="0.1",
#             creator="Cognite",
#             created=datetime.datetime.now(),
#             namespace=Namespace("http://purl.org/cognite/neat#"),
#             prefix="neat",
#             dataModelName="OpenAPI",
#             cdf_space_name="OpenAPI",
#             data_model_name="OpenAPI",
#         )

#         classes: dict[str, Class] = {}
#         properties: dict[str, Property] = {}

#         # Loop through OpenAPI components
#         for component_name, component_info in openapi_spec.get("components", {}).get("schemas", {}).items():
#             class_name = self.remove_underscores_and_capitalize(component_name)
#             class_name = self.get_dms_compatible_name(class_name)
#             class_id = class_name
#             logging.info(f" OpenAPi parser : Processing class {class_id} ")
#             try:
#                 class_ = Class(
#                     class_id=class_id,
#                     class_name=class_name,
#                     description=component_info.get("description", component_info.get("title", "empty")),
#                     is_abstract=False,
#                     is_interface=False,
#                 )
#                 classes[class_id] = class_
#                 # Loop through properties of OpenApi spec]
#                 self.process_properies(properties, class_id, class_name, component_info)
#             except Exception as e:
#                 logging.error(f" OpenAPi parser : Error creating class {class_id}: {e}")

#         rules = TransformationRules(
#             metadata=metadata, classes=classes, properties=properties, prefixes={}, instances=[]
#         )

#         return rules

#     def process_properies(
#         self,
#         rules_properties: dict[str, Property],
#         class_id: str,
#         class_name: str,
#         component: dict,
#         parent_property_name: str = None,
#     ):
#         # component can have keys : type, description, properties, required, title, allOf, anyOf, oneOf§
#         logging.info(f" OpenAPi parser : Processing properties for class {class_id} , component {component}")
#         for component_name, component_info in component.items():
#             if component_name == "allOf" or component_name == "anyOf" or component_name == "oneOf":
#                 for sub_component in component_info:
#                     self.process_properies(rules_properties, class_id, class_name, sub_component, component_name)
#             elif component_name == "properties":
#                 for prop_name, prop_info in component_info.items():
#                     prop_id = prop_name
#                     if prop_name == "allOf" or prop_name == "anyOf" or prop_name == "oneOf":
#                         if isinstance(prop_info, list):
#                             prop_type = prop_info[0].get("type", "string")
#                         else:
#                             logging.error(f" !!!!! prop info is not a list . its: {prop_info} ")

#                     else:
#                         prop_type = prop_info.get("type", "string")
#                         if prop_type == "array":
#                             self.process_properies(
#                                 rules_properties, class_id, class_name, prop_info.get("items", {}), prop_name
#                             )
#                             continue
#                     expected_value_type = self.map_open_api_type(prop_type)
#                     if prop_name == "$ref":
#                         ref_class = self.get_ref_class_name(prop_info.get("$ref", ""))
#                         expected_value_type = ref_class
#                     try:
#                         logging.info(f" OpenAPi parser : Processing property {prop_id} ")
#                         prop = Property(
#                             class_id=class_id,
#                             class_name=class_name,
#                             property_id=self.get_dms_compatible_name(prop_id),
#                             property_name=self.get_dms_compatible_name(prop_name.replace(".", "_")),
#                             property_type="ObjectProperty",
#                             description=prop_info.get("description", prop_info.get("title", "empty")),
#                             expected_value_type=expected_value_type,
#                             cdf_resource_type="Asset",
#                             resource_type_property="Asset",
#                             rule_type="rdfpath",
#                             rule=f"neat:{class_name}(neat:{prop_name})",
#                             label="linked to",
#                         )
#                         rules_properties[class_id + prop_id] = prop
#                     except Exception as e:
#                         logging.error(f" OpenAPi parser : Error creating property {prop_id}: {e}")
#             elif component_name == "$ref":
#                 ref_class = self.get_ref_class_name(component_info)
#                 logging.debug(f" OpenAPi parser : REF class {ref_class} ")
#                 prop = Property(
#                     class_id=class_id,
#                     class_name=class_name,
#                     property_id=parent_property_name,
#                     property_name=parent_property_name,
#                     property_type="ObjectProperty",
#                     description="no",
#                     expected_value_type=ref_class,
#                     cdf_resource_type="Asset",
#                     resource_type_property="Asset",
#                     rule_type="rdfpath",
#                     rule=f"neat:{class_name}(neat:{parent_property_name})",
#                     label="linked to",
#                 )
#                 rules_properties[class_id + parent_property_name] = prop

#     def get_ref_class_name(self, ref: str) -> str:
#         return self.get_dms_compatible_name(ref.split("/")[-1])

#     def get_dms_compatible_name(self, name: str) -> str:
#         # reserverd words in DMS
#         reserved_words_mapping = {
#             "space": "src_space",
#             "externalId": "external_id",
#             "createdTime": "created_time",
#             "lastUpdatedTime": "last_updated_time",
#             "deletedTime": "deleted_time",
#             "edge_id": "src_edge_id",
#             "node_id": "src_node_id",
#             "project_id": "src_project_id",
#             "property_group": "src_property_group",
#             "seq": "src_seq",
#             "tg_table_name": "src__table_name",
#             "extensions": "src_extensions",
#         }
#         if name in reserved_words_mapping:
#             return reserved_words_mapping[name]
#         else:
#             return name.replace(".", "_")

#     def remove_underscores_and_capitalize(self, input_string: str):
#         """Remove underscores and capitalize each word in the string , the convertion is done to improve complienc with DMS naming conventions"""
#         if "_" in input_string:
#             words = input_string.split("_")  # Split the string by underscores
#             result = "".join([word.capitalize() for word in words])  # Capitalize each word
#             return result
#         else:
#             return input_string

#     def map_open_api_type(self, openapi_type: str) -> str:
#         """Map OpenAPI type to NEAT compatible types"""
#         if openapi_type == "object":
#             datatype = "json"
#         elif openapi_type == "array":
#             datatype = "sequence"
#         elif openapi_type == "number":
#             datatype = "float"
#         else:
#             return openapi_type  # Default to string
#         return datatype
